//**************************************************************************** 
//
//  crt0.S (C Run Time 0)
//  
//  This is startup file for C main project.
//			
//  Extern definitions needed:
//    - UND_STACK_SIZE - undefined mode stack size
//    - ABT_STACK_SIZE - abort mode stack size
//    - FIQ_STACK_SIZE - fast interrupt mode stack size
//    - IRQ_STACK_SIZE - interrupt mode stack size
//    - SVC_STACK_SIZE - supervisor mode stack size
//
//  autor: Mariusz Janiak
//
//  Wroclaw 2006		
//
//****************************************************************************

// Standard definitions of Mode bits and Interrupt (I & F) flags in PSRs
        .set  MODE_USR, 0x10            // User Mode
        .set  MODE_FIQ, 0x11            // FIQ Mode
        .set  MODE_IRQ, 0x12            // IRQ Mode
        .set  MODE_SVC, 0x13            // Supervisor Mode
        .set  MODE_ABT, 0x17            // Abort Mode
        .set  MODE_UND, 0x1B            // Undefined Mode
        .set  MODE_SYS, 0x1F            // System Mode

        .equ  I_BIT, 0x80               // when I bit is set, IRQ is disabled
        .equ  F_BIT, 0x40               // when F bit is set, FIQ is disabled
        .equ  T_BIT, 0x20               // when T bit is set, THUMB mode

// VPBDIV definitions
        .equ    VPBDIV,      0xE01FC100  // VPBDIV Address 
        .equ    VPBDIV_Val,  0x02        // VPB divider value: 
                                         //  - 0x00 - 1/4 CCLK, 
                                         //  - 0x01 - 1/1 CCLK, 
                                         //  - 0x02 - 1/2 CCLK
// Phase Locked Loop (PLL) definitions
        .equ  PLL_BASE,      0xE01FC080  // PLL Base Address
        .equ  PLLCON_OFS,    0x00        // PLL Control Offset
        .equ  PLLCFG_OFS,    0x04        // PLL Configuration Offset 
        .equ  PLLSTAT_OFS,   0x08        // PLL Status Offset 
        .equ  PLLFEED_OFS,   0x0C        // PLL Feed Offset 
        .equ  PLLCON_PLLE,   (1<<0)      // PLL Enable 
        .equ  PLLCON_PLLC,   (1<<1)      // PLL Connect
        .equ  PLLSTAT_PLOCK, (1<<10)     // PLL Lock Status
        .equ  PLLCFG_PSEL,   (0x01<<5)	 // PSEL (P=2)
        .equ  PLLCFG_MSEL,   0x04        // MSEL (M=5)
        .equ  PLLCFG_Val,    (PLLCFG_PSEL|PLLCFG_MSEL)

// Memory Accelerator Module (MAM) definitions
        .equ  MAM_BASE,      0xE01FC000  // MAM Base Address 
        .equ  MAMCR_OFS,     0x00        // MAM Control Offset
        .equ  MAMTIM_OFS,    0x04        // MAM Timing Offset 
        .equ  MAMCR_Val,     0x00000002
        .equ  MAMTIM_Val,    0x00000004

// Memory Mapping (when Interrupt Vectors are in RAM)
        .equ  MEMMAP,        0xE01FC040  // Memory Mapping Control
		
		.text
        .arm

        .global _boot
        .func   _boot
_boot:

// Runtime Interrupt Vectors
// -------------------------
Vectors:
        ldr   pc,_reset                 // reset - _start
        ldr   pc,_undf                  // undefined - _undf
        ldr   pc,_swi                   // SWI - _swi
        ldr   pc,_pabt                  // program abort - _pabt
        ldr   pc,_dabt                  // data abort - _dabt
        nop                             // reserved
        ldr   pc,[pc,#-0xFF0]           // IRQ - read the VIC
        ldr   pc,_fiq                   // FIQ - _fiq

_reset:	.word _start	
_undf:  .word UNDEF_int                 // undefined
_swi:   .word SWI_int                   // SWI
_pabt:  .word PABT_int                  // program abort
_dabt:  .word DABT_int                  // data abort
        .word 0
_irq:   .word 0                         // handled by VIC        
_fiq:   .word FIQ_int                   // FIQ

        .size _boot, . - _boot
        .endfunc

// Pre-defined vectored interrupt handler SWI interrupt
// ----------------------------------------------------
        .global SWI_int
        .func   SWI_int

SWI_int:
		str	    ip, [sp, #-4]!
    	mov	    ip, sp
		stmdb	sp!, {r0-r3,fp,ip,lr,pc}
		sub	    fp, ip,#4
#if THUMB_INTERWORK
		mrs     r0,SPSR
        tst     r0,#T_BIT               // SWI call from 
        ldrneh  r0,[lr,#-2]             // Thumb: Load halfword instruction
        bicne   r0,r0,#0xFF00           //        extract SWI number
        ldreq   r0,[lr,#-4]             // ARM:   Load word instruction
        biceq   r0,r0,#0xFF000000       //        extract SWI number
#else
        ldr     r0,[lr,#-4]             // ARM:   Load word instruction
        bic     r0,r0,#0xFF000000       //        extract SWI number
#endif
		mov     r1,sp
		bl      SWI_sysCall
		sub	    sp, fp,#28
		ldmia	sp, {r0-r3,fp,sp,lr}
		ldmia	sp!, {ip}
		movs	pc, lr

		.size SWI_int, . - SWI_int
        .endfunc

// Setup the operating mode & stack.
// ---------------------------------
        .global _start, start
        .func   _start
_start:
start:

// VPB divider setup
// -----------------
#ifdef VPB_SETUP
        ldr   r0, =VPBDIV
        ldr   r1, =VPBDIV_Val
        str   r1, [r0]
#endif

// PLL setup
// ---------
#ifdef PLL_SETUP
        ldr   r0,=PLL_BASE
        mov   r1,#0xAA
        mov   r2,#0x55

// Configure and Enable PLL
        mov   r3,#PLLCFG_Val
        str   r3,[r0, #PLLCFG_OFS] 
        mov   r3,#PLLCON_PLLE
        str   r3,[r0, #PLLCON_OFS]
        str   r1,[r0, #PLLFEED_OFS]
        str   r2,[r0, #PLLFEED_OFS]

// Wait until PLL Locked
PLL_Loop:       
	ldr   r3,[r0, #PLLSTAT_OFS]
        ands  r3,r3, #PLLSTAT_PLOCK
        beq   PLL_Loop

// Switch to PLL Clock
        mov   r3,#(PLLCON_PLLE | PLLCON_PLLC)
        str   r3,[r0, #PLLCON_OFS]
        str   r1,[r0, #PLLFEED_OFS]
        str   r2,[r0, #PLLFEED_OFS]
#endif

// Memory Accelerator Module (MAM)
// -------------------------------
#ifdef MAM_SETUP
        ldr   r0,=MAM_BASE
        mov   r1,#MAMTIM_Val
        str   r1,[r0, #MAMTIM_OFS] 
        mov   r1,#MAMCR_Val
        str   r1,[r0, #MAMCR_OFS] 
#endif

// Memory Mapping (when Interrupt Vectors are in RAM)
// --------------------------------------------------	
#ifdef RAM_INTVEC
        ldr     r0,=MEMMAP
        mov     r1,#2
        str     r1,[r0]
#endif
	
// Initialize Interrupt System
// - Set stack location for each mode
// - Leave in System Mode with Interrupts Disabled
// -----------------------------------------------
        ldr   r0,=_stack_end
        msr   CPSR_c,#MODE_UND|I_BIT|F_BIT // Undefined Instruction Mode
        mov   sp,r0
        sub   r0,r0,#UND_STACK_SIZE
        msr   CPSR_c,#MODE_ABT|I_BIT|F_BIT // Abort Mode
        mov   sp,r0
        sub   r0,r0,#ABT_STACK_SIZE
        msr   CPSR_c,#MODE_FIQ|I_BIT|F_BIT // FIQ Mode
        mov   sp,r0
        sub   r0,r0,#FIQ_STACK_SIZE
        msr   CPSR_c,#MODE_IRQ|I_BIT|F_BIT // IRQ Mode
        mov   sp,r0
        sub   r0,r0,#IRQ_STACK_SIZE
        msr   CPSR_c,#MODE_SVC|I_BIT|F_BIT // Supervisor Mode
        mov   sp,r0
        sub   r0,r0,#SVC_STACK_SIZE
        msr   CPSR_c,#MODE_SYS|I_BIT|F_BIT // System Mode
        mov   sp,r0

// Copy initialized data to its execution address in RAM
// -----------------------------------------------------
        ldr   r1,=_etext                // -> ROM data start
        ldr   r2,=_data                 // -> data start
        ldr   r3,=_edata                // -> end of data
1:      cmp   r2,r3                     // check if data to move
        ldrlo r0,[r1],#4                // copy it
        strlo r0,[r2],#4
        blo   1b                        // loop until done
// Clear .bss
// ----------
        mov   r0,#0                     // get a zero
        ldr   r1,=__bss_start           // -> bss start
        ldr   r2,=__bss_end__           // -> bss end
2:      cmp   r1,r2                     // check if data to clear
        strlo r0,[r1],#4                // clear 4 bytes
        blo   2b                        // loop until done


// Call C++ constructors (for objects in "global scope")
// -----------------------------------------------------
#ifdef CPP_SUPORT
		ldr   r0, =__ctors_start__
		ldr   r1, =__ctors_end__
ctor_loop:
		cmp   r0, r1
		beq   ctor_end
		ldr   r2, [r0], #4
		stmfd sp!, {r0-r1}
		mov   lr, pc
		mov   pc, r2
		ldmfd sp!, {r0-r1}
		b 	  ctor_loop
ctor_end:
#endif
	
// Call main program: main(0)
// --------------------------
        mov   r0,#0                     // no arguments (argc = 0)
        mov   r1,r0
        mov   r2,r0
        mov   fp,r0                     // null frame pointer
        mov   r7,r0                     // null frame pointer for thumb
        adr   lr,_main_exit
	    b     main                      // enter main()

		
_main_exit:	
	b _main_exit
        
	.size   _start, . - _start
        .endfunc
        .end
